# -*- coding: utf-8 -*-
import pytest
from copy import copy
import sqlalchemy as sa
from sqlalchemy.engine.url import make_url

from cgstats.db import api

@pytest.fixture
def x_run_dir():
    return 'tests/fixtures/170202_ST-E00269_0169_AHC7H2ALXX'

@pytest.fixture
def x_pooled_run_dir():
    return 'tests/fixtures/161125_ST-E00269_0150_AH37GVALXX'

@pytest.fixture
def x_pooled_missing_logs_run_dir():
    return 'tests/fixtures/missing/all/161125_ST-E00269_0150_AH37GVALXX'

@pytest.fixture
def x_pooled_missing_unaligned_run_dir():
    return 'tests/fixtures/missing/Unaligned/161125_ST-E00269_0150_AH37GVALXX'

@pytest.fixture
def rapid_run_dir():
    return 'tests/fixtures/150114_D00134_0168_AHB07NADXX'

@pytest.fixture
def mixed_rapid_run_dir():
    return 'tests/fixtures/170406_D00410_0399_BHHKV5BCXY'

@pytest.fixture
def miseq_run_dir():
    return 'tests/fixtures/170609_M03284_0061_000000000-B59L9'

@pytest.fixture(scope="function")
def sql_manager(request):
    config_url = request.config.getoption('--sqlalchemy-connect-url')

    # get the databse from the URL
    sa_url = copy(make_url(config_url))
    database = sa_url.database
    sa_url.database = None

    engine = sa.create_engine(sa_url)
    engine.execute("CREATE DATABASE IF NOT EXISTS {}".format(database))

    manager = api.connect(config_url)
    manager.create_all()

    def fin():
        print("Removing database")
        manager.rollback()
        engine.execute("DROP DATABASE {}".format(database))
    
    request.addfinalizer(fin)
    return manager

def pytest_addoption(parser):
    parser.addoption("--sqlalchemy-connect-url", action="store",
                     default=None,
                     help="Name of the database to connect to")



# -*- coding: utf-8 -*-
import json

from sqlalchemy import Column, types, orm, ForeignKey

from cgstats.db import Model


class SampleFlowcell(Model):

    """Represent a single flowcell for a sample."""

    id = Column(types.Integer, primary_key=True)
    flowcell_id = Column(types.String(16))
    sample_id = Column(ForeignKey('analysis_sample.id'))

    lane = Column(types.Integer)
    cbot_machine = Column(types.String(32))
    sequencing_machine = Column(types.String(32))
    sequencing_date = Column(types.DateTime)
    run_mode = Column(types.String(32))
    sequencing_reagents_1 = Column(types.String(32))
    sequencing_reagents_2 = Column(types.String(32))
    sequencing_cluster_1 = Column(types.String(32))
    sequencing_cluster_2 = Column(types.String(32))
    cluster_concentration = Column(types.Float)
    q30 = Column(types.Float)
    reads = Column(types.Integer)


class AnalysisSample(Model):

    """Sample level QC metrics from an anlaysis."""

    id = Column(types.Integer, primary_key=True)
    sample_id = Column(types.String(32), unique=True)
    analysis_id = Column(ForeignKey('analysis.id'))
    sequencing_type = Column(types.Enum('wes', 'wgs'))
    sex_predicted = Column(types.Enum('male', 'female', 'unknown'))

    # pre-sequencing lab stuff
    capture_kit = Column(types.String(32))
    library_kit_lot_1 = Column(types.String(32))
    library_kit_lot_2 = Column(types.String(32))
    # can sometimes be a mix of two kits
    baits_lot = Column(types.String(64))
    beads_t1_lot = Column(types.String(32))
    fragment_size = Column(types.Integer)
    dna_concentration = Column(types.Float)
    input_material = Column(types.Float)
    index_sequence = Column(types.String(32))

    # mapping/alignment
    reads_total = Column(types.Integer)
    mapped_percent = Column(types.Float)
    duplicates_percent = Column(types.Float)
    strand_balance = Column(types.Float)

    # coverage
    coverage_target = Column(types.Float)
    completeness_target_10 = Column(types.Float)
    completeness_target_20 = Column(types.Float)
    completeness_target_50 = Column(types.Float)
    completeness_target_100 = Column(types.Float)

    # OMIM stats
    # 10-100X completeness

    # variant calling
    variants = Column(types.Integer)  # WHERE?
    indels = Column(types.Integer)
    snps = Column(types.Integer)
    novel_sites = Column(types.Integer)  # WHERE?
    concordant_rate = Column(types.Float)
    hethom_ratio = Column(types.Float)
    titv_ratio = Column(types.Float)

    flowcells = orm.relationship('SampleFlowcell', cascade='all,delete',
                                 backref='sample')

    @property
    def read_pairs(self):
        return self.reads_total / 2


class Analysis(Model):

    """Meta-data class to group samples from the same analysis."""

    id = Column(types.Integer, primary_key=True)
    # composed from "{customer_id}-{name}"
    analysis_id = Column(types.String(64), unique=True)
    pipeline = Column(types.String(32))
    pipeline_version = Column(types.String(32))
    analyzed_at = Column(types.DateTime)
    _program_versions = Column(types.Text)

    @property
    def program_versions(self):
        return json.loads(self._program_versions) if self._program_versions else {}

    @program_versions.setter
    def program_versions(self, value):
        self._program_versions = json.dumps(value)

    samples = orm.relationship('AnalysisSample', cascade='all,delete',
                               backref='analysis')

    @property
    def family_id(self):
        return self.analysis_id.split('-', 1)[-1]



# -*- coding: utf-8 -*-
from datetime import datetime
import json

from alchy import ModelBase, make_declarative_base
from sqlalchemy import Column, types, orm, ForeignKey, UniqueConstraint
from sqlalchemy.orm.exc import NoResultFound


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


class JsonModel(ModelBase):

    def to_json(self, pretty=False):
        """Serialize to JSON.

        Handle DateTime objects.
        """
        kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
        return json.dumps(self.to_dict(), default=json_serial, **kwargs)


Model = make_declarative_base(Base=JsonModel)


class Project(Model):

    project_id = Column(types.Integer, primary_key=True)
    projectname = Column(types.String(255), nullable=False)
    comment = Column(types.Text)
    time = Column(types.DateTime, nullable=False)

    samples = orm.relationship('Sample', cascade='all', backref=orm.backref('project'))

    def __repr__(self):
        return (u"{self.__class__.__name__}: {self.project_id}"
                .format(self=self))

    @staticmethod
    def exists(project_name):
        """Checks if the Prohect entry already exists

        Args:
            project_name (str): project name without the Project_ prefix

        Returns:
            int: project_id on exists
            False: on not exists

        """
        try:
            rs = (Project.query
                         .filter_by(projectname=project_name).one())
            return rs.project_id
        except NoResultFound:
            return False


class Sample(Model):

    sample_id = Column(types.Integer, primary_key=True)
    project_id = Column(ForeignKey('project.project_id', ondelete='CASCADE'), nullable=False)
    samplename = Column(types.String(255), nullable=False)
    customerid = Column(types.String(255))
    limsid = Column(types.String(255))
    barcode = Column(types.String(255))
    time = Column(types.DateTime)

    unaligned = orm.relationship('Unaligned', cascade='all, delete-orphan', backref=orm.backref('sample'))

    @property
    def lims_id(self):
        """Parse out the LIMS id from the samplename in demux database."""
        sample_part = self.samplename.split('_')[0]
        sanitized_id = sample_part.rstrip('FB')
        return sanitized_id

    @staticmethod
    def exists(sample_name, barcode):
        """Checks if a Sample entry already exists

        Args:
            sample_name (str): sample name without Sample_ prefix but with
                               index identifier _nxdual9
            barcode (str): the index

        Returns:
            int: sample_id on exists
            False: on not exists

        """
        try:
            rs = (Sample.query
                        .filter_by(samplename=sample_name)
                        .filter_by(barcode=barcode).one())
            return rs.sample_id
        except NoResultFound:
            return False


class Unaligned(Model):

    unaligned_id = Column(types.Integer, primary_key=True)
    sample_id = Column(ForeignKey('sample.sample_id', ondelete='CASCADE'), nullable=False)
    demux_id = Column(ForeignKey('demux.demux_id', ondelete='CASCADE'), nullable=False)
    lane = Column(types.Integer)
    yield_mb = Column(types.Integer)
    passed_filter_pct = Column(types.Numeric(10, 5))
    readcounts = Column(types.Integer)
    raw_clusters_per_lane_pct = Column(types.Numeric(10, 5))
    perfect_indexreads_pct = Column(types.Numeric(10, 5))
    q30_bases_pct = Column(types.Numeric(10, 5))
    mean_quality_score = Column(types.Numeric(10, 5))
    time = Column(types.DateTime)

    #demux = orm.relationship('Demux', backref=orm.backref('unaligned'))
    #sample = orm.relationship('Sample', single_parent=True, cascade='all, delete-orphan', passive_deletes=True, backref=orm.backref('unaligned'))

    @staticmethod
    def exists(sample_id, demux_id, lane):
        """Checks if an Unaligned entry already exists

        Args:
            sample_id (int): sample id
            demux_id (int): demux id
            lane (int): lane in which the sample ran

        Returns:
            int: unaligned_id on exists
            False: on not exists

        """
        try:
            rs = (Unaligned.query
                           .filter_by(sample_id=sample_id)
                           .filter_by(demux_id=demux_id)
                           .filter_by(lane=lane).one())
            return rs.unaligned_id
        except NoResultFound:
            return False


class Supportparams(Model):

    supportparams_id = Column(types.Integer, primary_key=True)
    document_path = Column(types.String(255), nullable=False, unique=True)
    systempid = Column(types.String(255))
    systemos = Column(types.String(255))
    systemperlv = Column(types.String(255))
    systemperlexe = Column(types.String(255))
    idstring = Column(types.String(255))
    program = Column(types.String(255))
    commandline = Column(types.Text)
    sampleconfig_path = Column(types.String(255))
    sampleconfig = Column(types.Text)
    time = Column(types.DateTime)

    def __repr__(self):
        return (u'{self.__class__.__name__}: {self.document_path}'.format(self=self))

    @staticmethod
    def exists(document_path):
        """Checks if the supportparams entry already exists

        Args:
            document_path (str): Full path to the Unaligned directory

        Returns:
            int: supportparams_id on exists
            False: on not exists

        """
        try:
            rs = (Supportparams.query
                               .filter_by(document_path=document_path).one())
            return rs.supportparams_id
        except NoResultFound:
            return False


class Datasource(Model):

    datasource_id = Column(types.Integer, primary_key=True)
    supportparams_id = Column(ForeignKey('supportparams.supportparams_id', ondelete='CASCADE'),
                              nullable=False)
    runname = Column(types.String(255))
    machine = Column(types.String(255))
    rundate = Column(types.Date)
    document_path = Column(types.String(255), nullable=False)
    document_type = Column(types.Enum('html', 'xml', 'undefined'),
                           nullable=False, default='html')
    server = Column(types.String(255))
    time = Column(types.DateTime)

    supportparams = orm.relationship('Supportparams', cascade='all',
                                     backref=orm.backref('datasources'))

    def __repr__(self):
        return (u'{self.__class__.__name__}: {self.runname}'.format(self=self))

    @staticmethod
    def exists(document_path):
        """Checks if the Datasource entry already exists

        Args:
            document_path (str): Full path to the stats file

        Returns:
            int: datasource_id on exists
            False: on not exists

        """
        try:
            rs = (Datasource.query
                            .filter_by(document_path=document_path).one())
            return rs.datasource_id
        except NoResultFound:
            return False


class Demux(Model):

    __table_args__ = (UniqueConstraint('flowcell_id', 'basemask',
                                       name='demux_ibuk_1'),)

    demux_id = Column(types.Integer, primary_key=True)
    flowcell_id = Column(ForeignKey('flowcell.flowcell_id', ondelete='CASCADE'), nullable=False)
    datasource_id = Column(ForeignKey('datasource.datasource_id', ondelete='CASCADE'), nullable=False)
    basemask = Column(types.String(255))
    time = Column(types.DateTime)

    datasource = orm.relationship('Datasource', cascade='all', backref=orm.backref('demuxes'))
    unaligned = orm.relation('Unaligned', single_parent=True, cascade='all, delete-orphan', backref=orm.backref('demux'))
    # add the viewonly attribute to make sure the Session.delete(demux) works
    samples = orm.relation('Sample', secondary='unaligned', viewonly=True, cascade='all', backref=orm.backref('demuxes'))

    @staticmethod
    def exists(flowcell_id, basemask):
        """Checks if the Demux entry already exists

        Args:
            flowcell_id (int): flowcell_id in the table Flowcell
            basemask (str): the basemask used to demux, e.g. Y101,I6n,Y101

        Returns:
        int: demux_id on exists
            False: on not exists

        """
        try:
            rs = (Demux.query
                       .filter_by(flowcell_id=flowcell_id)
                       .filter_by(basemask=basemask).one())
            return rs.demux_id
        except NoResultFound:
            return False


class Flowcell(Model):

    flowcell_id = Column(types.Integer, primary_key=True)
    flowcellname = Column(types.String(255), nullable=False, unique=True)
    flowcell_pos = Column(types.Enum('A', 'B'), nullable=False)
    hiseqtype = Column(types.String(255), nullable=False)
    time = Column(types.DateTime)

    demux = orm.relationship('Demux', cascade='all', backref=orm.backref('flowcell'))

    @staticmethod
    def exists(flowcell_name):
        """Checks if the Flowcell entry already exists

        Args:
            flowcell_name (str): The name of the flowcell

        Returns:
            int: flowcell_id on exists
            False: on not exists

        """
        try:
            rs = Flowcell.query.filter_by(flowcellname=flowcell_name).one()
            return rs.flowcell_id
        except NoResultFound:
            return False


class Backup(Model):

    runname = Column(types.String(255), primary_key=True)
    startdate = Column(types.Date, nullable=False)
    nas = Column(types.String(255))
    nasdir = Column(types.String(255))
    starttonas = Column(types.DateTime)
    endtonas = Column(types.DateTime)
    preproc = Column(types.String(255))
    preprocdir = Column(types.String(255))
    startpreproc = Column(types.DateTime)
    endpreproc = Column(types.DateTime)
    frompreproc = Column(types.DateTime)
    analysis = Column(types.String(255))
    analysisdir = Column(types.String(255))
    toanalysis = Column(types.DateTime)
    fromanalysis = Column(types.DateTime)
    inbackupdir = Column(types.Integer)
    backuptape_id = Column(ForeignKey('backuptape.backuptape_id'), nullable=False)
    backupdone = Column(types.DateTime)
    md5done = Column(types.DateTime)

    tape = orm.relationship('Backuptape', backref=orm.backref('backup'))

    @staticmethod
    def exists(runname, tapedir=None):
        """Check if run is already backed up. Optionally: checks if run is
        on certain tape

        Args:
            runname (str): e.g. 151117_D00410_0187_AHWYGMADXX
            tapedir (str): the name of the tape, e.g. tape036_037

        Returns:
            int: runname on exists
            False: on not exists
        """
        try:
            if tapedir is not None:
                rs = (Backup.query
                            .outerjoin(Backuptape)
                            .filter_by(runname=runname)
                            .filter(Backuptape.tapedir == tapedir).one())
            else:
                rs = Backup.query.filter_by(runname=runname).one()
            return rs.runname
        except NoResultFound:
            return False


class Backuptape(Model):

    backuptape_id = Column(types.Integer, primary_key=True)
    tapedir = Column(types.String(255))
    nametext = Column(types.String(255))
    tapedate = Column(types.DateTime)

    @staticmethod
    def exists(tapedir):
        """Check if a tape already exists

        Args:
            tapedir (str): the name of the tape, e.g. tape036_037

        Returns:
            int: backuptape_id on exists
            False: on not exists

        """
        try:
            rs = Backuptape.query.filter_by(tapedir=tapedir).one()
            return rs.backuptape_id
        except NoResultFound:
            return False


class Version(Model):

    __table_args__ = (UniqueConstraint('name', 'major', 'minor', 'patch',
                                       name='name_major_minor_patch_uc'),)

    version_id = Column(types.Integer, primary_key=True)
    name = Column(types.String(255))
    major = Column(types.Integer)
    minor = Column(types.Integer)
    patch = Column(types.Integer)
    comment = Column(types.String(255))
    time = Column(types.DateTime, nullable=False)

    @staticmethod
    def get_version():
        """Retrieves the database version

        Returns (tuple): (major, minor, patch, name)
        """

        """SELECT major, minor, patch, name FROM version ORDER BY time DESC LIMIT 1"""
        return Version.query.order_by(Version.time.desc()).limit(1).one()

    @staticmethod
    def check(dbname, ver):
        """Checks version of database against dbname and version

        [normally from the config file]

        Args:
          dbname (str): database name as stored in table version
          ver (str): version string in the format major.minor.patch

        Returns:
          True: if identical
        """
        rs = Version.get_version()
        if rs is not None:
            ver_string = "{0}.{1}.{2}".format(str(rs.major), str(rs.minor),
                                              str(rs.patch))
            return ((ver_string == ver) and (dbname == rs.name))



# -*- coding: utf-8 -*-
import datetime
from .models import Analysis, AnalysisSample

BEGINNING = datetime.datetime(1970, 1, 1)


def samples(query=None):
    """Return a list of samples."""
    query = query or AnalysisSample.query
    query = (query.join(AnalysisSample.analysis)
                  .order_by(Analysis.analyzed_at))
    return query


def duplicates():
    """Ask the database about duplicates."""
    results = {}
    for seq_type in ('wes', 'wgs'):
        query = samples().filter(AnalysisSample.sequencing_type == seq_type)
        percentages = [{
            'name': sample.sample_id,
            'y': (sample.duplicates_percent * 100)
        } for sample in query]
        results[seq_type] = percentages
    return results


def readsvscov(db):
    """Ask the database about reads vs. coverage."""
    results = {}
    for seq_type in ('wes', 'wgs'):
        query = db.query(AnalysisSample.reads_total,
                         AnalysisSample.completeness_target_20 * 100)
        query = (samples(query=query).
                 filter(AnalysisSample.sequencing_type == seq_type,
                        AnalysisSample.reads_total != None))
        results[seq_type] = [(row[0], round(row[1], 1)) for row in query]
    return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import socket
import pytest
from path import Path

from cgstats.db.xparse import gather_supportparams, gather_datasource, gather_demux

def test_gather_supportparams(x_run_dir, x_pooled_missing_logs_run_dir, x_pooled_missing_unaligned_run_dir):

    with pytest.raises(IOError):
        gather_supportparams(x_pooled_missing_logs_run_dir)

    with pytest.raises(IOError):
        gather_supportparams(x_pooled_missing_unaligned_run_dir)

    assert gather_supportparams(x_run_dir) == {
            'idstring': 'bcl2fastq v2.15.0.4',
            'commandline': '/usr/local/bin/bcl2fastq -d 2 -r 4 -w 4 -p 14 --tiles s_1_11 --tiles s_1_12 -R /scratch/743545//mnt/hds2/proj/bioinfo/Runs/170202_ST-E00269_0169_AHC7H2ALXX -o /scratch/743545/Xout --barcode-mismatches 1 --use-bases-mask Y151,I8,Y151',
            'time': '20170205061756',
            'program': '/usr/local/bin/bcl2fastq',
            'sampleconfig_path': str(Path(x_run_dir).joinpath('SampleSheet.csv')),
            'sampleconfig': '[Data]\n'
'FCID,Lane,SampleID,SampleRef,index,SampleName,Control,Recipe,Operator,Project\n'
'HC7H2ALXX,1,SVE2274A2_TCCGCGAA,hg19,TCCGCGAA,659262,N,R1,NN,659262\n'
'HC7H2ALXX,2,SVE2274A4_TCCGCGAA,hg19,TCCGCGAA,659262,N,R1,NN,659262\n'
'HC7H2ALXX,3,SVE2274A6_TCCGCGAA,hg19,TCCGCGAA,659262,N,R1,NN,659262\n'
'HC7H2ALXX,4,SVE2274A7_TCCGCGAA,hg19,TCCGCGAA,659262,N,R1,NN,659262\n'
'HC7H2ALXX,5,SVE2274A8_TCCGCGAA,hg19,TCCGCGAA,659262,N,R1,NN,659262\n'
'HC7H2ALXX,6,SVE2274A9_TCTCGCGC,hg19,TCTCGCGC,659262,N,R1,NN,659262\n'
'HC7H2ALXX,7,SVE2274A10_TCTCGCGC,hg19,TCTCGCGC,659262,N,R1,NN,659262\n'
'HC7H2ALXX,8,SVE2274A11_TCTCGCGC,hg19,TCTCGCGC,659262,N,R1,NN,659262',
            'document_path': str(Path(x_run_dir).joinpath('Unaligned'))
    }

def test_gather_datasource(x_run_dir, x_pooled_missing_logs_run_dir):

    with pytest.raises(IOError):
        gather_datasource(x_pooled_missing_logs_run_dir)

    assert gather_datasource(x_run_dir) == {
            'runname': str(Path(x_run_dir).normpath().basename()),
            'rundate': '170202',
            'machine': 'ST-E00269',
            'servername': socket.gethostname(),
            'document_path': str(Path(x_run_dir).joinpath('l1t11/Stats/ConversionStats.xml'))
    }

def test_gather_demux(x_run_dir, x_pooled_missing_logs_run_dir):

    with pytest.raises(IOError):
        gather_demux(x_pooled_missing_logs_run_dir)

    assert gather_demux(x_run_dir) == {
        'basemask': 'Y151,I8,Y151'
    }



#!/usr/bin/env python
# encoding: utf-8

from __future__ import print_function, division
import sys
import os
from glob import glob
import logging
import socket

from path import Path
from sqlalchemy import func

from demux.utils import Samplesheet
from cgstats.db.models import Supportparams, Version, Datasource, Flowcell, Demux, Project, Sample, Unaligned
from cgstats.utils import xstats
from cgstats.utils.utils import get_projects, gather_flowcell

logger = logging.getLogger(__name__)


class NoLogsFoundError(Exception):
    pass


def gather_supportparams(run_dir):
    """Aggregates all the support params:
    - bcl2fastq version
    - bcl2fastq path
    - command executed
    - datetime
    - SampleSheet.csv path
    - SampleSheet
    - DEMUX path

    Args:
        run_dir (str): FQPN run dir

    Returns: dict(
        'document_path',
        'idstring',
        'program',
        'commandline',
        'sampleconfig_path',
        'sampleconfig',
        'time')
    """

    rs = {} # result set

    # get some info from bcl2 fastq
    run_dir = Path(run_dir)
    logfile = run_dir.joinpath('LOG', 'Xdem-l?t??-*.log')
    logfilenames = glob(logfile) # should yield one result
    if len(logfilenames) == 0:
        logger.error('No log files found! Looking for {}'.format(logfile))
        import errno
        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), logfile)

    with open(logfilenames[0], 'r') as logfile:
        for line in logfile.readlines():
            if 'bcl2fastq v' in line:
                rs['idstring'] = line.strip()

            if '--use-bases-mask' in line:
                line = line.strip()
                split_line = line.split(' ')
                rs['commandline'] = ' '.join(split_line[1:]) # remove the leading [date]
                rs['time'] = split_line[0].strip('[]') # remove the brackets around the date
                rs['program'] = split_line[1] # get the executed program

    # get the sample sheet and it's contents
    samplesheet_path = run_dir.joinpath('SampleSheet.csv')
    rs['sampleconfig_path'] = str(samplesheet_path)
    rs['sampleconfig'] = Samplesheet(samplesheet_path).raw()

    # get the unaligned dir
    document_path = run_dir.joinpath('Unaligned')
    if not os.path.isdir(document_path):
        logger.error("Unaligned dir not found at '{}'".format(document_path))
        import errno
        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), document_path)
    else:
        rs['document_path'] = str(document_path)

    return rs

def gather_datasource(run_dir):
    """TODO: Docstring for gather_datasource.

    Args:
        run_dir (TODO): TODO

    Returns: TODO

    """

    run_dir = Path(run_dir)
    rs = {} # result set

    # get the run name
    rs['runname'] = str(run_dir.normpath().basename())

    # get the run date
    rs['rundate'] = rs['runname'].split('_')[0]

    # get the machine name
    rs['machine'] = rs['runname'].split('_')[1]

    # get the server name on which the demux took place
    rs['servername'] = socket.gethostname()

    # get the stats file
    document_path = run_dir.joinpath('l1t11/Stats/ConversionStats.xml')
    if not document_path.isfile():
        logger.error("Stats file not found at '%s'", document_path)
        import errno
        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), document_path)
    else:
        rs['document_path'] = str(document_path)

    return rs

def gather_demux(run_dir):
    """TODO: Docstring for gather_demux.

    Args:
        demux_dir (TODO): TODO

    Returns: TODO

    """

    rs = {} # result set

    # get some info from bcl2 fastq
    run_dir = Path(run_dir)
    logfile = run_dir.joinpath('LOG', 'Xdem-l?t??-*.log')
    logfilenames = glob(logfile) # should yield one result
    if len(logfilenames) == 0:
        logger.error('No log files found! Looking for {}'.format(logfile))
        import errno
        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), logfile)

    with open(logfilenames[0], 'r') as logfile:
        for line in logfile.readlines():

            if '--use-bases-mask' in line:
                line = line.strip()
                split_line = line.split(' ')
                basemask_params_pos = [i for i,x in enumerate(split_line) if x == '--use-bases-mask'][0]
                rs['basemask'] = split_line[basemask_params_pos + 1]

    return rs

def sanitize_sample(sample):
    """Removes the _nxdual9 index indication
    Removes the B (reprep) or F (reception fail) suffix from the sample name

    Args:
        sample (str): a sample name

    Return (str): a sanitized sample name

    """
    return sample.split('_')[0].rstrip('BF')

def get_sample_sheet(demux_dir):
    sample_sheet = []
    samplesheet_file_name = glob("{demux_dir}/SampleSheet.csv".format(demux_dir=demux_dir))[0]
    with open(samplesheet_file_name, 'r') as samplesheet_fh:
        lines = [ line.strip().split(',') for line in samplesheet_fh.readlines() ]
        header = []
        for line in lines:
            # skip headers
            if line[0].startswith('['): continue
            if line[2] == 'SampleID':
                header = line
                continue

            entry = dict(zip(header, line))
            sample_sheet.append(entry)

    return sample_sheet

def get_nr_samples_lane(sample_sheet):
    samples_lane = {}
    for line in sample_sheet:
        if line['Lane'] not in samples_lane:
             samples_lane[ line['Lane'] ] = 0
        samples_lane[ line['Lane'] ] += 1

    return samples_lane

def add(manager, demux_dir):
    """ Gathers and adds all data to cgstats.

    params:
        manager (managerAlchamy): a manager object which can be used to query the DB
        demux_dir (str): the demux dir of the run
    returns: true on success!
    """

    # ok, let's process the support params
    supportparams_id = Supportparams.exists(os.path.join(demux_dir, 'Unaligned'))
    if not supportparams_id:
        new_supportparams = gather_supportparams(demux_dir)
        supportparams = Supportparams()
        supportparams.document_path = new_supportparams['document_path']
        supportparams.idstring = new_supportparams['idstring']
        supportparams.program = new_supportparams['program']
        supportparams.commandline = new_supportparams['commandline']
        supportparams.sampleconfig_path = new_supportparams['sampleconfig_path']
        supportparams.sampleconfig = new_supportparams['sampleconfig']
        supportparams.time = new_supportparams['time']

        manager.add(supportparams)
        manager.flush()
        supportparams_id = supportparams.supportparams_id

    datasource_id = Datasource.exists(os.path.join(demux_dir, 'l1t11/Stats/ConversionStats.xml'))
    if not datasource_id:
        new_datasource = gather_datasource(demux_dir)
        datasource = Datasource()
        datasource.runname = new_datasource['runname']
        datasource.rundate = new_datasource['rundate']
        datasource.machine = new_datasource['machine']
        datasource.server = new_datasource['servername']
        datasource.document_path = new_datasource['document_path']
        datasource.document_type = 'xml'
        datasource.time = func.now()
        datasource.supportparams_id = supportparams_id

        manager.add(datasource)
        manager.flush()
        datasource_id = datasource.datasource_id

    flowcell_namepos = gather_flowcell(demux_dir)
    flowcell_name = flowcell_namepos['name']
    flowcell_pos  = flowcell_namepos['pos']
    flowcell_id   = Flowcell.exists(flowcell_name)
    if not flowcell_id:
        flowcell = Flowcell()
        flowcell.flowcellname = flowcell_name
        flowcell.flowcell_pos = flowcell_pos
        flowcell.hiseqtype = 'hiseqx'
        flowcell.time = func.now()

        manager.add(flowcell)
        manager.flush()
        flowcell_id = flowcell.flowcell_id

    new_demux = gather_demux(demux_dir)
    demux_id = Demux.exists(flowcell_id, new_demux['basemask'])
    if not demux_id:
        demux = Demux()
        demux.flowcell_id = flowcell_id
        demux.datasource_id = datasource_id
        demux.basemask = new_demux['basemask']
        demux.time = func.now()

        manager.add(demux)
        manager.flush()
        demux_id = demux.demux_id

    project_id_of = {} # project name: project id
    for project_name in get_projects(demux_dir):
        project_id = Project.exists(project_name)
        if not project_id:
            p = Project()
            p.projectname = project_name
            p.time = func.now()

            manager.add(p)
            manager.flush()
            project_id = p.project_id

        project_id_of[ project_name ] = project_id

    sample_sheet = get_sample_sheet(demux_dir)
    stats = xstats.parse(demux_dir)
    stats_samples = xstats.parse_samples(demux_dir)
    nr_samples_lane = get_nr_samples_lane(sample_sheet)
    for sample in sample_sheet:
        sample_id = Sample.exists(sample['SampleID'], sample['index'])
        if not sample_id:
            s = Sample()
            s.project_id = project_id_of[ sample['Project'] ]
            s.samplename = sample['SampleID']
            s.limsid = sample['SampleID'].split('_')[0]
            s.barcode = sample['index']
            s.time = func.now()

            manager.add(s)
            manager.flush()
            sample_id = s.sample_id

        if not Unaligned.exists(sample_id, demux_id, sample['Lane']):
            u = Unaligned()
            u.sample_id = sample_id
            u.demux_id = demux_id
            u.lane = sample['Lane']
            if nr_samples_lane[ sample['Lane'] ] > 1: # pooled!
                stats_sample = stats_samples[ sample['Lane'] ][ sample['SampleID'] ]

                u.yield_mb = round(int(stats_sample['pf_yield']) / 1000000, 2)
                u.passed_filter_pct = stats_sample['pf_yield_pc']
                u.readcounts = stats_sample['pf_clusters'] * 2
                u.raw_clusters_per_lane_pct = stats_sample['raw_clusters_pc']
                u.perfect_indexreads_pct = round(stats_sample['perfect_barcodes'] / stats_sample['barcodes'] * 100, 5)
                u.q30_bases_pct = stats_sample['pf_Q30']
                u.mean_quality_score = stats_sample['pf_qscore']
            else:
                u.yield_mb = round(int(stats[ sample['SampleID'] ]['pf_yield']) / 1000000, 2)
                u.passed_filter_pct = stats[ sample['SampleID'] ]['pf_yield_pc']
                u.readcounts = stats[ sample['SampleID'] ]['pf_clusters'] * 2
                u.raw_clusters_per_lane_pct = stats[ sample['SampleID'] ]['raw_clusters_pc']
                u.perfect_indexreads_pct = round(stats[ sample['SampleID'] ]['perfect_barcodes'] / stats[ sample['SampleID'] ]['barcodes'] * 100, 5)
                u.q30_bases_pct = stats[ sample['SampleID'] ]['pf_Q30']
                u.mean_quality_score = stats[ sample['SampleID'] ]['pf_qscore']
            u.time = func.now()

            manager.add(u)

    manager.flush()
    manager.commit()

    return True

if __name__ == '__main__':
    xadd(sys.argv[1:])



# -*- coding: utf-8 -*-

import yaml
from os.path import expanduser

class BaseConfig:

    def __init__(self):
        with open(expanduser("~/.clinical/databases.yaml"), 'r') as ymlfile:
            self.config = yaml.load(ymlfile)

    def __getitem__(self, key):
        """Simple array-based getter.

        Args:
            key (str): Gets the value for this key in the YAML file.

        Returns (str, dict): returns the structure underneat this key.

        """
        return self.config[key]



# -*- coding: utf-8 -*-
import logging
import pkg_resources

from alchy import Manager
from sqlalchemy import func

from .models import Model, Sample, Flowcell, Demux, Datasource, Unaligned, Project
from sqlalchemy import or_

SAMPLE_PATTERN = "{}\_%"
log = logging.getLogger(__name__)


def connect(uri):
    """Connect to the database."""
    for models in pkg_resources.iter_entry_points('cgstats.models.1'):
        models.load()
    log.debug('open connection to database: %s', uri)
    manager = Manager(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=Model,
                      session_options=dict(autoflush=False))
    return manager


def get_sample(sample_id):
    """Get a unique demux sample."""
    pattern = SAMPLE_PATTERN.format(sample_id)
    query = Sample.query.filter(or_(Sample.samplename.like(pattern),
                                    Sample.samplename == sample_id))
    return query


def flowcells(sample=None):
    """Return a query for the latest flowcells."""
    query = (Flowcell.query.join(Demux).join(Datasource)
                           .order_by(Datasource.rundate.desc()))
    if sample:
        pattern = SAMPLE_PATTERN.format(sample)
        query = (query.join(Demux.unaligned, Unaligned.sample)
                      .filter(or_(Sample.samplename.like(pattern),
                                  Sample.samplename == sample)))
    return query


def samples(flowcell_name=None):
    """Return a query for the latest samples."""
    query = (Sample.query.join(Sample.unaligned, Unaligned.demux,
                               Demux.flowcell))
    if flowcell_name:
        query = query.filter(Flowcell.flowcellname == flowcell_name)
    return query


def select(flowcell, project_name=None):
    """Return a query for the stats"""

#    query = """ SELECT sample.samplename AS smp, flowcell.flowcellname AS flc, 
#    GROUP_CONCAT(unaligned.lane ORDER BY unaligned.lane) AS lanes, 
#    GROUP_CONCAT(unaligned.readcounts ORDER BY unaligned.lane) AS rds, SUM(unaligned.readcounts) AS readsum, 
#    GROUP_CONCAT(unaligned.yield_mb ORDER BY unaligned.lane) AS yield, SUM(unaligned.yield_mb) AS yieldsum, 
#    GROUP_CONCAT(TRUNCATE(q30_bases_pct,2) ORDER BY unaligned.lane) AS q30, 
#    GROUP_CONCAT(TRUNCATE(mean_quality_score,2) ORDER BY unaligned.lane) AS meanq
#    FROM sample, flowcell, unaligned, project, demux
#    WHERE sample.sample_id = unaligned.sample_id 
#    AND flowcell.flowcell_id = demux.flowcell_id
#    AND unaligned.demux_id = demux.demux_id 
#    AND sample.project_id = project.project_id
#    AND project.projectname = '""" + proje + """' 
#    AND flowcell.flowcellname = '""" + flowc + """'
#    GROUP BY samplename, flowcell.flowcell_id 
#    ORDER BY lane, sample.samplename, flowcellname """

    query = (Sample.query.join(Sample.unaligned, Unaligned.demux, Demux.flowcell))
    if project_name:
        query = (query.join(Sample.project).filter(Project.projectname == project_name))

    query = query.with_entities(
                  Sample.samplename, Flowcell.flowcellname,
                  func.group_concat(Unaligned.lane.op('ORDER BY')(Unaligned.lane)).label('lanes'),
                  func.group_concat(
                      Unaligned.readcounts.op('ORDER BY')(Unaligned.lane)).label('reads'),
                  func.sum(Unaligned.readcounts).label('readsum'),
                  func.group_concat(
                      Unaligned.yield_mb.op('ORDER BY')(Unaligned.lane)).label('yld'),
                  func.sum(Unaligned.yield_mb).label('yieldsum'),
                  func.group_concat(
                      func.truncate(Unaligned.q30_bases_pct, 2)
                      .op('ORDER BY')(Unaligned.lane)).label('q30'),
                  func.group_concat(
                      func.truncate(Unaligned.mean_quality_score, 2)
                      .op('ORDER BY')(Unaligned.lane)).label('meanq'),
            )

    query = query.filter(Flowcell.flowcellname == flowcell)
    query = query.group_by(Sample.samplename, Flowcell.flowcell_id)
    query = query.order_by(Unaligned.lane, Sample.samplename, Flowcell.flowcellname)

    return query




# -*- coding: utf-8 -*-
from __future__ import division
import os

from flask import Flask, render_template
from flask_alchy import Alchy
from flask_bootstrap import Bootstrap

from cgstats.db import Model
from cgstats.analysis import api as analysis_api
from cgstats.analysis.models import AnalysisSample

TEMPLATES_AUTO_RELOAD = True
SQLALCHEMY_TRACK_MODIFICATIONS = False
SQLALCHEMY_DATABASE_URI = os.environ['SQLALCHEMY_DATABASE_URI']
if 'mysql' in SQLALCHEMY_DATABASE_URI:  # pragma: no cover
    SQLALCHEMY_POOL_RECYCLE = 3600

app = Flask(__name__)
app.config.from_object(__name__)
db = Alchy(app, Model=Model)
Bootstrap(app)


@app.template_filter()
def millions(value):
    return "{} M".format(int(value / 1000000))


@app.template_filter()
def percent(value):
    return round(value * 100, 1)


@app.route('/')
def index():
    """Dashboard view."""
    most_dups = (AnalysisSample.query.order_by(AnalysisSample.duplicates_percent.desc())
                               .limit(5))
    least_mapped = (AnalysisSample.query.filter(AnalysisSample.mapped_percent != None)
                                  .order_by(AnalysisSample.mapped_percent)
                                  .limit(5))
    order_attr = AnalysisSample.completeness_target_10
    filter_cond = AnalysisSample.completeness_target_10 != None
    least_complete = (AnalysisSample.query.filter(filter_cond)
                                    .order_by(order_attr).limit(5))
    dups = analysis_api.duplicates()
    readsvscov = analysis_api.readsvscov(db)
    return render_template('index.html', dups=dups, readsvscov=readsvscov,
                           most_dups=most_dups, least_mapped=least_mapped,
                           least_complete=least_complete)


@app.route('/samples')
def samples():
    """Show raw samples data."""
    samples_q = analysis_api.samples().limit(50)
    return render_template('samples.html', samples=samples_q)


@app.route('/report')
def report():
    """Show qc report."""
    samples_q = analysis_api.samples().limit(5)
    return render_template('report.html', samples=samples_q)



#!/usr/bin/env python
# encoding: utf-8

from __future__ import print_function, division

from bs4 import BeautifulSoup

def parse(demux_stats):
    """parses the Demultiplex_Stats.htm

    Args:
        demux_stats (str): a path to Demultiplex_Stats.htm of the run

    Returns: { lane: { sample_id: {}}}

    """

    def mkint(s):
        """Make sure empty strings default to 0"""
        return int('0'+s)

    def mkfloat(s):
        """Make sure empty strings default to 0"""
        return float('0'+s)

    samples = {} # lane: sample_id: {}

    soup = BeautifulSoup(open(demux_stats), 'html.parser')
    tables = soup.findAll("table")
    rows = tables[1].findAll('tr')
    for row in rows:
        sample = {}
        cols = row.findAll('td')

        lane = cols[0].text
        if not lane in samples:
            samples[lane] = {}

        sample_name = cols[1].text
        sample['sample_name'] = sample_name
        sample['barcode'] = cols[3].text
        sample['project_id'] = cols[6].text
        sample['lane'] = lane
        sample['yield_mb'] = mkint(cols[7].text.replace(",",""))
        sample['pf_pc'] = mkfloat(cols[8].text)
        sample['readcounts'] = mkint(cols[9].text.replace(",",""))
        sample['raw_clusters_pc'] = mkfloat(cols[10].text)
        sample['perfect_barcodes_pc'] = mkfloat(cols[11].text)
        sample['q30_bases_pc'] = mkfloat(cols[13].text)
        sample['mean_quality_score'] = mkfloat(cols[14].text)

        samples[lane][sample_name] = sample

    return samples
